VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Diagram"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
'****************************************************
'* Diagram.cls                                      *
'* class representing a diagram                     *
'* Programmed: Konrad Rudolph (Mad Rat Pro.)        *
'* Last Change: 01.02.2004 19:29                    *
'* Version: 1.0.0                                   *
'****************************************************
Option Explicit

'----------------------------------------------------
' private members
'----------------------------------------------------

Private WithEvents Canvas   As PictureBox       ' picturebox to paint into
Attribute Canvas.VB_VarHelpID = -1

Private m_ValuesFetched As Boolean              ' have the values been fetched?
Private m_Values(QUARTS_PER_DAY - 1) As Long    ' table of all our values
Private m_StartTime     As Date                 ' stating time of the current interval

Private m_ColBG         As OLE_COLOR            ' background colour of graph
Private m_ColGrid       As OLE_COLOR            ' grid lines colour
Private m_StlGrid       As DrawStyleConstants   ' grid lines style
Private m_ColLine       As OLE_COLOR            ' graph line colour
Private m_StlLine       As DrawStyleConstants   ' graph line style
Private m_ColText       As OLE_COLOR            ' colour of axis captions
Private m_FntText       As StdFont              ' font of the axis captions
Private m_ColHigh       As OLE_COLOR            ' colour of highlighted lines and captions
Private m_StlHigh       As DrawStyleConstants   ' style of highlighted lines
Private m_ColVariance   As OLE_COLOR            ' colour of variance graph
Private m_StlVariance   As DrawStyleConstants   ' style of variance graph
Private m_DoHighlight   As Boolean

Private m_BasePath      As String               ' path to the data files

Private m_ScaleY        As Long                 ' number of captions on y axis

Private m_MarginLeft    As Long                 ' left margin
Private m_MarginBottom  As Long                 ' bottom margin
'----------------------------------------------------

'----------------------------------------------------
' a great many public properties ...
'----------------------------------------------------

Public Property Get Target() As PictureBox
    Set Target = Canvas
End Property

Public Property Set Target(RHS As PictureBox)
    Set Canvas = RHS
End Property

Public Property Get BackColor() As OLE_COLOR
    BackColor = m_ColBG
End Property
Public Property Let BackColor(RHS As OLE_COLOR)
    m_ColBG = RHS
End Property

Public Property Get GridColor() As OLE_COLOR
    GridColor = m_ColGrid
End Property
Public Property Let GridColor(RHS As OLE_COLOR)
    m_ColGrid = RHS
End Property

Public Property Get GridStyle() As DrawStyleConstants
    GridStyle = m_StlGrid
End Property
Public Property Let GridStyle(RHS As DrawStyleConstants)
    m_StlGrid = RHS
End Property

Public Property Get LineColor() As OLE_COLOR
    LineColor = m_ColLine
End Property
Public Property Let LineColor(RHS As OLE_COLOR)
    m_ColLine = RHS
End Property

Public Property Get LineStyle() As DrawStyleConstants
    LineStyle = m_StlLine
End Property
Public Property Let LineStyle(RHS As DrawStyleConstants)
    m_StlLine = RHS
End Property

Public Property Get TextColor() As OLE_COLOR
    TextColor = m_ColText
End Property
Public Property Let TextColor(RHS As OLE_COLOR)
    m_ColText = RHS
End Property

Public Property Get Font() As StdFont
    Set Font = m_FntText
End Property
Public Property Set Font(RHS As StdFont)
    Set m_FntText = RHS
End Property

Public Property Get HighlightColor() As OLE_COLOR
    HighlightColor = m_ColHigh
End Property
Public Property Let HighlightColor(RHS As OLE_COLOR)
    m_ColHigh = RHS
End Property

Public Property Get HighlightStyle() As DrawStyleConstants
    HighlightStyle = m_StlHigh
End Property
Public Property Let HighlightStyle(RHS As DrawStyleConstants)
    m_StlHigh = RHS
End Property

Public Property Get VarianceColor() As OLE_COLOR
    VarianceColor = m_ColVariance
End Property
Public Property Let VarianceColor(RHS As OLE_COLOR)
    m_ColVariance = RHS
End Property

Public Property Get VarianceStyle() As DrawStyleConstants
    VarianceStyle = m_StlVariance
End Property
Public Property Let VarianceStyle(RHS As DrawStyleConstants)
    m_StlVariance = RHS
End Property

Public Property Get UseHighlight() As Boolean
    UseHighlight = m_DoHighlight
End Property
Public Property Let UseHighlight(RHS As Boolean)
    m_DoHighlight = RHS
End Property

Public Property Get Basepath() As String
    Basepath = m_BasePath
End Property
Public Property Let Basepath(RHS As String)
    m_BasePath = RHS
End Property

Public Property Get ScaleY() As Long
    ScaleY = m_ScaleY
End Property
Public Property Let ScaleY(RHS As Long)
    m_ScaleY = RHS
End Property
'----------------------------------------------------

'----------------------------------------------------
' class constructor
'----------------------------------------------------

Private Sub Class_Initialize()
    ' assign standard values
    
    m_ColBG = vbWhite
    m_ColGrid = &HD0D0D0
    m_ColLine = vbBlack
    m_ColText = vbBlack
    m_ColHigh = vbBlack
    m_ColVariance = vbRed
    
    m_DoHighlight = False
    
    Set m_FntText = New StdFont
    m_FntText.Name = "Tahoma"
    m_FntText.Size = 10
    
    m_BasePath = App.Path
    If Right$(m_BasePath, 1) <> "\" Then _
        m_BasePath = m_BasePath & "\"
    
    m_ScaleY = 5            ' this is the number given in the assignment;
                            ' personally, I would recommend a higher value: 10 or near
End Sub
'----------------------------------------------------

'----------------------------------------------------
' private object event handlers
'----------------------------------------------------

Private Sub Canvas_Resize()
    Call Refresh
End Sub

Private Sub Canvas_MouseMove(Button As Integer, Shift As Integer, X As Single, Y As Single)
    Static LastX As Long    ' prevents flickering
    
    Dim MyX As Long, _
        MyY As Long
    Dim Dat As String       ' values to write
    
    If m_ValuesFetched And Not (Canvas Is Nothing) Then
        If m_DoHighlight Then
            ' highlights the value hovered by cursor:
            ' draw a highlight crosshair and the exact values
            
            ' get nearest x
            
            MyX = Int(X + 0.5) - m_MarginLeft
            
            ' if x is in valid range and not different from last x ...
            
            If MyX >= 0 And MyX < QUARTS_PER_DAY And MyX <> LastX Then
                MyY = m_Values(MyX) ' assign y
                
                Canvas.Cls
                Canvas.DrawStyle = m_StlHigh
                
                If MyY > -1 Then ' y is a valid value
                    ' draw crosshair
                    
                    Canvas.Line ( _
                        MyX + m_MarginLeft, 0 _
                        )-( _
                        MyX + m_MarginLeft, Canvas.ScaleHeight - m_MarginBottom _
                        ), m_ColHigh
                    Canvas.Line ( _
                        m_MarginLeft, Canvas.ScaleHeight - MyY - m_MarginBottom _
                        )-( _
                        Canvas.ScaleWidth, Canvas.ScaleHeight - MyY - m_MarginBottom _
                        ), m_ColHigh
                    
                    ' draw values:
                    
                    Canvas.ForeColor = m_ColHigh
                    
                    ' time ...
                    
                    ' format time for output
                    Dat = Format$( _
                        m_StartTime + _
                        TimeSerial(MyX \ QUARTS_PER_HOUR, (MyX Mod QUARTS_PER_HOUR) * 15, 0), "h:nn" _
                    )
                    Canvas.CurrentX = MyX + m_MarginLeft
                    Canvas.CurrentY = Canvas.ScaleHeight - m_MarginBottom
                    Canvas.Print CStr(Dat)
                    
                    ' ... value
                    
                    ' format value
                    Dat = Format$(MyY, "###,##0")
                    
                    ' align at the axis
                    Canvas.CurrentX = Canvas.TextWidth("000.000") - Canvas.TextWidth(Dat)
                    Canvas.CurrentY = Canvas.ScaleHeight - MyY - m_MarginBottom * 1.5
                    Canvas.Print CStr(Dat)
                    
                    LastX = MyX
                End If
            End If
        End If
    End If
End Sub
'----------------------------------------------------

'----------------------------------------------------
' public procedures
'----------------------------------------------------

'----------------------------------------------------
' Create
' creates a new graph
' - [IN] [EndDate]: specifies the ending of the interval
'                   to be displayed
'                   default value is current time and date
' - returns success
'----------------------------------------------------
Public Function Create(Optional EndDate As Variant) As Boolean
    Dim StartTime   As Date, _
        EndTime     As Date     ' interval boundaries
    Dim DataSource  As TextFile ' text file object
    Dim ValuesY()   As Long     ' day before
    Dim ValuesT()   As Long     ' to-day
    
    On Error GoTo ERR_Create
    
    ' determine interval
    
    If IsMissing(EndDate) Then _
        EndTime = Now _
    Else _
        EndTime = CDate(EndDate)
    StartTime = DateAdd("d", TimeSerial(-24, 0, 0), EndTime)
    
    ' round date for axis captions on graph
    
    m_StartTime = TimeSerial( _
        DatePart("h", StartTime), _
        DatePart("n", StartTime) \ 15 * 15, _
        0 _
    )
    
    Set DataSource = New TextFile
    
    ' read yesterday's values
    
    If Not DataSource.Read(m_BasePath, StartTime) Then _
        Exit Function   ' error reading file
    
    ValuesY = DataSource.Values
    
    ' read today's values
    
    If Not DataSource.Read(m_BasePath, EndTime) Then _
        Exit Function   ' error reading file
    
    ValuesT = DataSource.Values
    
    ' fetch values in the chosen interval
    
    Call GetInterval24(StartTime, EndTime, ValuesY, ValuesT)
    
    ' display graph
    
    Call Refresh
    
    Create = True
    
    On Error GoTo 0
    Exit Function

ERR_Create:
    Call Err.Raise( _
        vbObjectError, _
        "Diagram:Create()", _
        "Unhandled exception occurred" _
    )
End Function

'----------------------------------------------------
' Variance
' draws a variance graph for the current graph
'----------------------------------------------------
Public Sub Variance()
    Dim X               As Long     ' current x position
    Dim Diff            As Long     ' holds the current differentiated value
    Dim LastWasValid    As Boolean  ' determines whether line is to continue
                                    ' or to be restarted (after missing value)
    
    If m_ValuesFetched And Not (Canvas Is Nothing) Then
        Canvas.Cls
        Canvas.AutoRedraw = True
        
        Canvas.DrawStyle = m_StlVariance
        
        ' draw orientation line: x-axis for variance graph
        
        Canvas.Line ( _
            m_MarginLeft, Canvas.ScaleHeight / 2 - m_MarginBottom _
        )-( _
            Canvas.ScaleWidth, Canvas.ScaleHeight / 2 - m_MarginBottom _
        ), m_ColVariance
        
        For X = 1 To QUARTS_PER_DAY - 2
            ' do only if interval can be differentiated
            
            If m_Values(X - 1) > -1 And m_Values(X + 1) > -1 Then
                ' differentiate
                
                Diff = m_Values(X - 1) - m_Values(X + 1)
                
                If Not LastWasValid Then
                    ' reassign line starting position
                    
                    Canvas.CurrentX = X + m_MarginLeft
                    Canvas.CurrentY = Canvas.ScaleHeight / 2 + Diff - m_MarginBottom
                    LastWasValid = True
                End If
                
                Canvas.Line -( _
                    X + m_MarginLeft, Canvas.ScaleHeight / 2 + Diff - m_MarginBottom _
                ), m_ColVariance
            Else
                LastWasValid = False
            End If
        Next X
        
        Set Canvas.Picture = Canvas.Image
        Canvas.AutoRedraw = False
    End If
End Sub
'----------------------------------------------------

'----------------------------------------------------
' private procedures
'----------------------------------------------------

'----------------------------------------------------
' Refresh
' refreshes the graph
'----------------------------------------------------
Private Sub Refresh()
    If m_ValuesFetched And Not (Canvas Is Nothing) Then
        Canvas.AutoRedraw = True
        
        Set Canvas.Font = m_FntText
        Call DrawGrid
        Call DrawValues
        
        Set Canvas.Picture = Canvas.Image
        Canvas.AutoRedraw = False
    End If
End Sub

'----------------------------------------------------
' GetInterval24
' collects the values of the 24h interval
' - [IN] StartTime As Date: starting time of interval
' - [IN] EndTime As Date: ending time of interval
' - [IN] Yesterday() As Long: values of day before
' - [IN] Today() As Long: values of current day
'----------------------------------------------------
Private Sub GetInterval24( _
  StartTime As Date, _
  EndTime As Date, _
  Yesterday() As Long, _
  Today() As Long _
)
    Dim TimeToTime  As Long
    Dim Cnt         As Long
    Dim Index       As Long
    
    ' retrieve starting and ending index of the value arrays
    
    TimeToTime = GetIndexFromTime(StartTime)
    
    ' fill in values from our interval
    
    ' loop though values of day before
    
    For Cnt = TimeToTime To QUARTS_PER_DAY - 1
        m_Values(Index) = Yesterday(Cnt)
        Index = Index + 1
    Next Cnt
    
    ' loop through values of current day
    
    For Cnt = 0 To TimeToTime - 1
        m_Values(Index) = Today(Cnt)
        Index = Index + 1
    Next Cnt
    
    m_ValuesFetched = True
End Sub

'----------------------------------------------------
' DrawGrid
' draws the graph's grid, axis and the axis captions
'----------------------------------------------------
Private Sub DrawGrid()
    Dim X       As Long, _
        Y       As Long     ' current drawing position
    Dim DatTime As String   ' \ marking of the x (and y) axis
    Dim StrTime As String   ' /
    Dim Highest As Long     ' highes of all values (rounded to next higher significant value)
    
    Highest = HighestValue()
    
    With Canvas
        ' assign width and height
        
        .ScaleWidth = QUARTS_PER_DAY
        .ScaleHeight = Highest
        
        ' calculate margins for axis caption
        
        m_MarginLeft = .TextWidth("0.000.000")
        m_MarginBottom = .TextHeight("00:00")
        
        ' readjust width and height taking into account the margins
        
        .ScaleWidth = .ScaleWidth + m_MarginLeft
        .ScaleHeight = .ScaleHeight + m_MarginBottom
        
        ' assign font colour
        
        .ForeColor = m_ColText
        
        ' draw background
        
        .Cls
        Canvas.Line (0, 0)-(.ScaleWidth, .ScaleHeight), m_ColBG, BF
        
        ' draw grid:
        
        ' vertically ...
        
        .DrawStyle = m_StlGrid
        
        For X = 0 To QUARTS_PER_DAY - 1
            Canvas.Line (X + m_MarginLeft, 0)-(X + m_MarginLeft, .ScaleHeight - m_MarginBottom), m_ColGrid
        Next X
        
        ' ... and horizontally
        
        For Y = 0 To Highest - 1 Step (Highest \ m_ScaleY)
            Canvas.Line ( _
                m_MarginLeft, .ScaleHeight - Y - m_MarginBottom _
                )-( _
                .ScaleWidth, .ScaleHeight - Y - m_MarginBottom _
                ), m_ColGrid
        Next Y
        
        ' print axis values
        
        '   on x-axis ...
        
        .DrawStyle = vbSolid
        
        For X = 0 To QUARTS_PER_DAY Step (QUARTS_PER_HOUR * 2)
            ' this is to calculate the current time
            
            DatTime = m_StartTime + TimeSerial(X \ QUARTS_PER_HOUR, (X Mod QUARTS_PER_HOUR) * 15, 0)
            
            ' format time for output
            
            StrTime = Format$(DatTime, "h:nn")
            Canvas.CurrentX = X + m_MarginLeft
            Canvas.CurrentY = .ScaleHeight - m_MarginBottom
            Canvas.Print CStr(StrTime)
            
            Canvas.Line (X + m_MarginLeft, 0)-(X + m_MarginLeft, .ScaleHeight - m_MarginBottom), m_ColText
        Next X
        
        '   ... and on y-axis.
        
        For Y = 0 To Highest - 1 Step (Highest \ m_ScaleY)
            ' format value
            
            StrTime = Format$(Y, "###,##0")
            
            ' align at the axis
            
            Canvas.CurrentX = .TextWidth("000.000") - .TextWidth(StrTime)
            Canvas.CurrentY = .ScaleHeight - Y - m_MarginBottom * 1.5
            Canvas.Print CStr(StrTime)
        Next Y
        
        ' re-draw x-axis line
        
        Canvas.Line ( _
            m_MarginLeft, .ScaleHeight - m_MarginBottom _
            )-( _
            .ScaleWidth, .ScaleHeight - m_MarginBottom _
            ), m_ColText
    End With
End Sub

'----------------------------------------------------
' DrawValues
' draws the graph itself
'----------------------------------------------------
Private Sub DrawValues()
    Dim X               As Long     ' current x value
    Dim LastWasValid    As Boolean  ' determines whether line is to continue
                                    ' or to be restarted (after missing value)
    With Canvas
        .DrawStyle = m_StlLine
        
        For X = 0 To QUARTS_PER_DAY - 1
            If m_Values(X) > -1 Then
                If Not LastWasValid Then
                    ' reassign line starting position
                    
                    .CurrentX = X + m_MarginLeft
                    .CurrentY = .ScaleHeight - m_Values(X) - m_MarginBottom
                    LastWasValid = True
                End If
                
                Canvas.Line -(X + m_MarginLeft, .ScaleHeight - m_Values(X) - m_MarginBottom), m_ColLine
            Else
                LastWasValid = False
            End If
        Next X
        
    End With
End Sub

'----------------------------------------------------
' HighestValue As Long
' determines the highest value the interval
' this value is then rounded up to the next highest
' significant value (significance = number of digits - 2)
' - returns the value
'----------------------------------------------------
Private Function HighestValue() As Long
    Dim I               As Long ' current value's index
    Dim Significance    As Long ' significant places
    
    ' retrieve highest value
    
    For I = 0 To QUARTS_PER_DAY - 1
        If m_Values(I) > HighestValue Then _
            HighestValue = m_Values(I)
    Next I
    
    ' get significance of number
    ' significance =  number of digits - 2 = (logarithm base 10 of n) - 2
    
    Significance = Log(HighestValue) / Log(10) - 2
    
    ' round value to next higher
    
    If Significance > 0 Then _
        HighestValue = Ceil(HighestValue, Significance)
End Function
